package org.jgroups.protocols;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import org.jgroups.Address;
import org.jgroups.Event;
import org.jgroups.Message;
import org.jgroups.View;
import org.jgroups.debug.Simulator;
import org.jgroups.protocols.FD_SOCK;
import org.jgroups.protocols.FD_SOCK.FdHeader;
import org.jgroups.stack.IpAddress;
import org.jgroups.stack.Protocol;
import org.jgroups.util.DefaultThreadFactory;
import org.jgroups.util.QueueClosedException;
import org.jgroups.util.TimeScheduler;
import org.jgroups.util.Util;

import java.util.Properties;
import java.util.Vector;
import java.net.InetAddress;
import java.net.ServerSocket;


/**
 * Tests the FD_SOCK protocol
 * @author Dennis Reed
 */
public class FD_SOCKTest extends TestCase {
    IpAddress a1;
    IpAddress a2;
    Vector members;
    View v;
    Simulator s;
    ServerSocket sock;
    FD_SOCK fd_sock;
    Harness harness;

    public FD_SOCKTest(String name) {
        super(name);
    }

    public void setUp() throws Exception {
    	System.out.println("START");
        super.setUp();

        a1=new IpAddress(1234);
        a2=new IpAddress(1235);

        // View must contain at least 2 members for testJGRP-1888
        members=new Vector();
        members.add(a1);
        members.add(a2);
        v=new View(a1, 1, members);

        fd_sock=new FD_SOCK();
        Properties props=new Properties();
        props.setProperty("suspect_msg_interval", "100");
        fd_sock.setProperties(props);

        harness = new Harness();
        
        Protocol[] stack=new Protocol[]{fd_sock, harness};

        s=new Simulator();
        s.setLocalAddress(a1);
        s.setView(v);
        s.addMember(a1);
        s.setProtocolStack(stack);
        s.start();
    }

    public void tearDown() throws Exception {
        super.tearDown();
        s.stop();
    }

    /*
     * This triggers JGRP-1888 by:
     * 1. Tell byteman to pause in FD_SOCK#stop just after the bcast_task.removeAll() call
     * 2. Fake the other JBoss instance with a socket listener that doesn't do 
     *    anything.  We only need the connections opened for this test)
     */
    public void testJGRP1888() throws Exception {
    	harness.reset();
    	sock = new ServerSocket(2000, 1, InetAddress.getLoopbackAddress());
    	
    	// Give FD_SOCK time to get the socket address and connect to it
		Thread.sleep(1000);

		// Stop FD_SOCK
		new Thread ( new Runnable () {
			public void run () {
				fd_sock.stop();
			}
		} ).start ();

		// Wait for the thread to be blocked by byteman
		Thread.sleep ( 100 );

		// Stop the fake FD_SOCK listener
		sock.close ();

		// Wait.  Only one SUSPECT event should occur here
		Thread.sleep(1000);

		assertTrue("Expected one SUSPECT event but got " + harness.getSuspectCount(), harness.getSuspectCount() <= 1);
		System.out.println ( harness.getSuspectCount() + " events");
    }

    public class Harness extends Protocol
    {
    	private int suspectCount = 0;
    	
		@Override
		public Object down ( Event evt )
		{
			if ( evt.getType() == Event.MSG )
			{
				Message msg=(Message) evt.getArg();
				FdHeader hdr=(FdHeader)msg.getHeader("FD_SOCK");
				if(hdr != null) {
					switch(hdr.type) {
					case FdHeader.WHO_HAS_SOCK:
						sendWhoHasSockReply();
						break;
					case FdHeader.SUSPECT:
						suspectCount++;
						break;
					}
				}
			}
			return super.down(evt);
		}

		public void reset()
		{
			suspectCount = 0;
		}

		public int getSuspectCount()
		{
			return suspectCount;
		}

		public void sendWhoHasSockReply()
		{
			if(sock == null)
				return;
			Message msg=new Message(a1, a2, null);
			FdHeader hdr=new FdHeader(FdHeader.I_HAVE_SOCK);
			hdr.mbr=a2;
			hdr.sock_addr=new IpAddress(sock.getInetAddress(), sock.getLocalPort());
			msg.putHeader("FD_SOCK", hdr);
			up_prot.up(new Event(Event.MSG, msg));
		}
		
		@Override
		public String getName ()
		{
			return "Harness";
		}		
    }

    public static Test suite() {
        return new TestSuite(FD_SOCKTest.class);
    }

    public static void main(String[] args) {
        junit.textui.TestRunner.run(suite());
    }
}
